/*
 * Simple port test program.
 * The lpd program opens a socket/port to this program,
 * which then listens for the connection and reads from the
 * socket until done.
 */
/******************************************************************
 * System Include Files
 ******************************************************************/
#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/dir.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sgtty.h>
#include <ctype.h>
#ifdef USE_STRINGS
#include <strings.h>
#else USE_STRINGS
#include <string.h>
#endif USE_STRINGS
#include <errno.h>
#include <grp.h>


/******************************************************************
 * Declare Standard Routines.
 * Note: most of these are not declared in include files.
 * This leads to some interesting problems with LINT. Declare these
 * here.  Some specialized routines are declared in the individual
 * routines that use them.
 ******************************************************************/
extern char	*malloc(), *realloc();
extern char	*getenv();
extern char	*gets();
extern int errno, sys_nerr;
extern char *sys_errlist[];
extern char *index(), *rindex();
extern char *sprintf();	/* sigh... sprintf strikes again */
extern long lseek(), atol();

int  local_port = 2048;
char *Name;

main(argc, argv)
	int argc;
	char **argv;
{
	int f;						/* Acme Integer, Inc. strikes again */
	int finet;					/* Internet socket */
	int defreadfds, readfds;	/* select()ed connections */
	int	req;					/* request on this socket */
	struct sockaddr_in sock_in, frominet; /* networking addresses */
	int fromlen;
	int nfds;					/* accept return value */

	/*
	 * process command line options
	 */
	Name = argv[0];
	if( argc > 1 ){
		local_port = atoi( argv[1] );
	}

	log( 1,  "listen: port %d", local_port );
	/*
	 * Set up the IPC.  This is very ugly,  and is stolen directly from
	 * the original LPD code and the 4.3 BSD Interprocess Communication
	 * Tutorial, which seems to have borrowed it from RLOGIN, etc.
	 *
	 * Set up INET socket
	 */
	if( ( finet = socket(AF_INET, SOCK_STREAM, 0) ) < 0 ){
		logerr_die( 1, "listen: cannot create AF_INET socket");
	}
	/*
	 * zero the sockaddr structure before using it
	 */
	bzero( (char *)&sock_in, sizeof(sock_in) );
	/*
	 * the INADDR_ANY will allow different nets to be used
	 */
	sock_in.sin_family = AF_INET;
	sock_in.sin_addr.s_addr = INADDR_ANY;
	sock_in.sin_port = local_port;
	if (bind(finet, (struct sockaddr *)&sock_in, sizeof(sock_in)) < 0) {
		logerr_die( 1, "listen: bind failed internet domain");
	} 
	if( listen(finet, 5) < 0 ){
		logerr_die( 1, "listen: listen failed internet domain" );
	}
	log( 1,  "listen: AF_INET socket %d", finet );

	/*
	 * Set up the sockets on which we will do an accept
	 */
	defreadfds = (1 << finet);

	/*
	 * Main loop: accept, do a request, continue.
	 */
	for (;;) {
		log( 1, "listen: starting select");

		readfds = defreadfds;
		nfds = select(20, (fd_set *)&readfds, (fd_set *)0, (fd_set *)0,
			(struct timeval *)0); /* wait patiently */
		if (nfds <= 0) {
			if (nfds < 0 && errno != EINTR) {
				logerr_die( 1, "listen: select error");
			}
			continue;
		}
		/*
		 * incredible, but true: could have NO requests
		 */
		if( readfds ){
			log( 1, "listen: doing select 0x%x", readfds );
			/* get the socket causing the select */
			fromlen = sizeof(frominet);
			req = accept(finet, (struct sockaddr *)&frominet, &fromlen);
			if (req < 0) {
				logerr( 1, "listen: finet accept");
				continue;
			}
			if (!From_host(&frominet)){
				logerr( 1, "listen: From_host failed");
				continue;
			}
			Read_input(req);
			(void) close(req);
		}
	}
}
/***********************************************************************
 * From_host( sockaddr_in *f )
 *   This routine will extract the remote Host Name and remote Host
 *   port number information.  The remote process must connect on
 *   a port in a "reserved" range;  these are reserved for use
 *   by priviledged processes (setuid/uid root).
 * Side Effects: sets the "from" string to the Name of the remote Host.
 * Returns: 1 if privileged and successful, 0 if not.
 * NOTE: this code was derived from the orignal 4.2BSD LPR source,
 *   the 4.3BSD Interprocess Communications Tutorials, and other places.
 ***********************************************************************/

From_host(f)
	struct sockaddr_in *f;
{
	struct hostent *hp;			/* Host entry */
	extern char *inet_ntoa();	/* inet address translation routine */
	int port;					/* from port number */

	if (f->sin_family != AF_INET ){
		log( 1, "From_host: malformed from address, family %d, not AF_INET %d",
			f->sin_family, AF_INET);
		return(0);
	}
	port = ntohs(f->sin_port);
	hp = gethostbyaddr((char *)(&f->sin_addr), sizeof(struct in_addr),
				f->sin_family);
	if (hp == 0){
		logerr( 1, "From_host: Host Name for address '%s' unknown",
			inet_ntoa(f->sin_addr));
		return(0);
	}
	log( 1, "From_host: host '%s', port %d", hp->h_name, port );
	return(1);
}

/*
 * Errormsg( err ) returns a printable form of the errormessage
 * corresponding to errno.
 */

char *
Errormsg( err )
	int err;
{
	char *cp;
	static char msgbuf[30];

	if( err >= 0 && err <= sys_nerr ){
		cp = sys_errlist[err];
	} else {
		(void)sprintf(msgbuf,"errno=%d",err);
		cp = msgbuf;
	}
	return(cp);
}

/*
 * log( int severity; char *msg; args )
 *  	1. format the message; msg and args are similar to printf
 *		2. output on stderr
 *		3. if "Echo_on_stdout" output on stdout
 *		4. if severity high enough,  log using syslog()
 *		5. if not a Debug message, save in Last_errormsg[]
 */
/*VARARGS2*/
log(kind, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
	char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	char buf[BUFSIZ];
	int omask;

	buf[0] = 0;
	if(Name && *Name){
		(void)sprintf(buf+strlen(buf),"%s: ", Name);
	}
	(void)sprintf(buf+strlen(buf), msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);
	(void)fprintf(stderr,"%s\n", buf);
	(void)fflush(stderr);
}

/*VARARGS2*/
fatal(kind, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
	char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	log( kind, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);
	exit(1);
}

/*VARARGS2*/
logerr( kind,msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
	char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	int err = errno;
	char buf1[BUFSIZ];

	(void)sprintf(buf1, msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);
	log( kind,"%s - %s", buf1, Errormsg(err) );
	errno = err;
}

/*VARARGS2*/
logerr_die( kind,msg, a1, a2, a3, a4, a5, a6, a7, a8, a9)
	char *msg, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9;
{
	logerr( kind,msg, a1, a2, a3, a4, a5, a6, a7, a8, a9);
	exit(1);
}
